from typing import List, Union, Tuple, Dict

from sohstationviewer.view.util.enums import LogType
from sohstationviewer.conf import constants
from sohstationviewer.controller.util import get_val
from sohstationviewer.view.util.color import clr

# TODO: put this in DB
mass_pos_volt_ranges = {"regular": [0.5, 2.0, 4.0, 7.0],
                        "trillium": [0.5, 1.8, 2.4, 3.5]}
mass_pos_color_pallets = {"B": ["C", "G", "Y", "R", "M"],
                          "W": ["B", "B", "B", "B", "B"]}


def get_masspos_value_colors(
        range_opt: str, chan_id: str, c_mode: str,
        processing_log: List[Tuple[str, LogType]],
        ret_type: str = 'str'
) -> Union[str, List[Tuple[float, str]], None]:
    """
    Create a map between value and color based on given rangeOpt and c_mode to
    display mass position plots.
    :param range_opt: massPosVoltRangeOpt got from Options Menu - MP coloring
        in Main Window to define different values for mass position.
        (regular/trillium)
    :param chan_id: ID of the channel
    :param c_mode: color mode (B/W)
    :param processing_log: list of processing info and type
    :param ret_type: request return type
    :return: [(value, color), (value, color) ...]
        if retType is 'str', return "value:color|value:color"
    """

    if range_opt.lower() not in mass_pos_volt_ranges.keys():
        processing_log.append(
            (
                f"{chan_id}: The current selected Mass Position color range is"
                f" '{range_opt}' isn't allowable. The accept ranges are: "
                f"{', '.join(mass_pos_volt_ranges.keys())}",
                LogType.ERROR
            )
        )
        return
    mass_pos_volt_range = mass_pos_volt_ranges[range_opt]
    mass_pos_color_pallet = mass_pos_color_pallets[c_mode]
    value_colors = []
    for i in range(len(mass_pos_volt_range)):
        if ret_type == 'str':
            value_colors.append(
                "%s:%s" % (mass_pos_volt_range[i], mass_pos_color_pallet[i]))
        else:
            value_colors.append(
                (mass_pos_volt_range[i], mass_pos_color_pallet[i])
            )
        if i == len(mass_pos_volt_range) - 1:
            if ret_type == 'str':
                value_colors.append(
                    "%s:+%s" % (
                        mass_pos_volt_range[i], mass_pos_color_pallet[i + 1]))
            else:
                value_colors.append(
                    (mass_pos_volt_range[i], mass_pos_color_pallet[i + 1]))
    if ret_type == 'str':
        return '|'.join(value_colors)
    return value_colors


def get_categorized_data_from_value_color_equal_on_upper_bound(
        c_data: Dict, chan_db_info: Dict
) -> Tuple[List[List[float]], List[str]]:
    """
    Separate data points and color using valueColors in which the condition
    will check if value equal to or less than upper bound and greater than
    lower bound as following example
        0:R|2.3:Y|+2.3:G  means
            + value <= 0   => plot with R color
            + 0 < value <= 2.3 => plot with Y color
            + 2.3 < value  => plot with G color
    Beside additional signs are included:
        Ex: *:W  means everything with white color
        Ex: -1:_ means not plotting value <= -1
    :param c_data: dict of data of the channel which includes down-sampled
        data in keys 'times' and 'data'.
    :param chan_db_info: dict of info of channel from DB
    :return x: list of x to be plotted to get total samples
    :return colors: list of color to be plotted to decide color of total sample
        text.
    """
    prev_val = -constants.HIGHEST_INT
    value_colors = chan_db_info['valueColors'].split('|')
    colors = []
    points_list = []
    for vc in value_colors:
        v, c = vc.split(':')
        if v == '*':
            val = v         # to have some value for pre_val = val
        else:
            val = get_val(v)
        if c == '_':
            prev_val = val
            continue
        colors.append(c)
        # c_data['times'] is a list of one list (after changing to faster data
        # reading) so don't need to loop through but only taking the first
        # element (c_data['times'][0]).
        # Same for c_data['data']
        times, data = c_data['times'][0], c_data['data'][0]
        if v.startswith('+'):
            points = [times[i]
                      for i in range(len(data))
                      if data[i] > val]
        elif v == '*':
            points = times
        else:
            points = [times[i]
                      for i in range(len(data))
                      if prev_val < data[i] <= val]
        points_list.append(points)
        prev_val = val
    return points_list, colors


def get_categorized_data_from_value_color_equal_on_lower_bound(
        c_data: Dict, chan_db_info: Dict
) -> Tuple[List[List[float]], List[str]]:
    """
    Separate data points and color using valueColors in which the condition
    will check if value equal to or greater than lower bound and less than
    upper bound as the following example:
        3.:R|3.3:Y|=3.3:G  means
            + value < 3.   => plot with R color
            + 3<= value < 3.3 => plot with Y color
            + value = 3.3  => plot with G color
    :param c_data: dict of data of the channel which includes down-sampled
        data in keys 'times' and 'data'.
    :param chan_db_info: dict of info of channel from DB
    :return x: list of x to be plotted to get total samples
    :return colors: list of color to be plotted to decide color of total sample
        text.
    """
    prev_val = 0
    value_colors = chan_db_info['valueColors'].split('|')
    colors = []
    points_list = []
    for vc in value_colors:
        v, c = vc.split(':')
        colors.append(c)
        val = get_val(v)
        times, data = c_data['times'][0], c_data['data'][0]
        if v.startswith('='):
            points = [times[i]
                      for i in range(len(data))
                      if data[i] >= val]
        else:
            points = [times[i]
                      for i in range(len(data))
                      if prev_val <= data[i] < val]
        points_list.append(points)
        prev_val = val
    return points_list, colors


def get_colors_sizes_for_abs_y_from_value_colors(
        y: List[float],
        value_colors: List[Tuple[float, str]]) -> \
        Tuple[List[str], List[float]]:
    """
    Map each abs(value) in y with the value in value_colors to get the colors
    for the indexes corresponding to the y's items.
    Sizes is currently similar for all items
    :param y: list of data values
    :param value_colors: list of color with indexes corresponding to y's items
    :return colors: list of colors of markers corresponding to y
    :return sizes: list of sizes of markers corresponding to y
    """
    colors = [None] * len(y)
    sizes = [1.5] * len(y)
    for i in range(len(y)):
        count = -1
        for v, c in value_colors:
            count += 1
            if count <= len(value_colors) - 2:
                if abs(y[i]) <= v:
                    colors[i] = clr[c]
                    break
            else:
                # The last value color
                colors[i] = clr[c]
    return colors, sizes
